#include "ComponentCodeGenerator.h"

#include <iostream>
#include <fstream>
#include <set>
#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/dom/DOM.hpp>
#include <xercesc/sax/HandlerBase.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/XMLStringTokenizer.hpp>
#include <xercesc/util/PlatformUtils.hpp>

using namespace X3D;
using namespace std;

XERCES_CPP_NAMESPACE_USE

struct ltstr
{
  bool operator()(const SFString &s1, const SFString &s2) const
  {
    return strcmp(s1, s2) < 0;
  }
};

void ComponentCodeGenerator::generateFilesFromFile(const SFString &file)
{
  try 
  {
    XMLPlatformUtils::Initialize();
  }
  catch (const XMLException& toCatch) 
  {
    char* message = XMLString::transcode(toCatch.getMessage());
    cout << "Error during initialization! :\n"
         << message << "\n";
    XMLString::release(&message);
  }

  XercesDOMParser* parser = new XercesDOMParser();
  
  parser->setValidationScheme(XercesDOMParser::Val_Never);    // optional.
  parser->setDoNamespaces(false);
  parser->setDoSchema(false);
  parser->setValidationSchemaFullChecking(false);
    
  const char* xmlFile = file;

  try 
  {
    parser->parse(xmlFile);
  }
  catch (...) 
  {
    cout << "Unexpected Exception \n" ;
  }
  
  SFString module;
  MFString components;
  DOMElement *root = parser->getDocument()->getDocumentElement();
  DOMNode *child = root->getFirstChild();

  while (child != 0)
  {
    if (child->getNodeType() == DOMNode::ELEMENT_NODE)
    {
      if (SFString(XMLString::transcode(((DOMElement *)child)->getTagName())) == "Module")
      {
        module = XMLString::transcode(((DOMElement *)child)->getAttribute(XMLString::transcode("name")));
      }
      if (SFString(XMLString::transcode(((DOMElement *)child)->getTagName())) == "Components")
      {
        DOMNode *com = child->getFirstChild();
        while (com != 0)
        {
          if (com->getNodeType() == DOMNode::ELEMENT_NODE)
          {
            MFString enterList;
            MFString walkOnList;
            MFString leaveList;
            SFString component = XMLString::transcode(((DOMElement *)com)->getTagName());
            components.push_back(component);
            DOMNode *functions = com->getFirstChild();
            while (functions != 0)
            {
              if (functions->getNodeType() == DOMNode::ELEMENT_NODE)
              {
                if (SFString(XMLString::transcode(((DOMElement *)functions)->getTagName())) == "Enter")
                {
                  DOMNode *names = functions->getFirstChild();
                  if ((names != 0) && (names->getNodeType() == DOMNode::TEXT_NODE))
                  {
                    XMLStringTokenizer st(((DOMText *)names)->getNodeValue());
                    while (st.hasMoreTokens())
                      enterList.push_back(XMLString::transcode(st.nextToken()));
                  }
                }
                if (SFString(XMLString::transcode(((DOMElement *)functions)->getTagName())) == "WalkOn")
                {
                  DOMNode *names = functions->getFirstChild();
                  if ((names != 0) && (names->getNodeType() == DOMNode::TEXT_NODE))
                  {
                    XMLStringTokenizer st(((DOMText *)names)->getNodeValue());
                    while (st.hasMoreTokens())
                      walkOnList.push_back(XMLString::transcode(st.nextToken()));
                  }
                }
                if (SFString(XMLString::transcode(((DOMElement *)functions)->getTagName())) == "Leave")
                {
                  DOMNode *names = functions->getFirstChild();
                  if ((names != 0) && (names->getNodeType() == DOMNode::TEXT_NODE))
                  {
                    XMLStringTokenizer st(((DOMText *)names)->getNodeValue());
                    while (st.hasMoreTokens())
                      leaveList.push_back(XMLString::transcode(st.nextToken()));
                  }
                }
              }
              functions = functions->getNextSibling();
            }

            //generate files
            generateFiles(module, component, enterList, walkOnList, leaveList);
          }
          com = com->getNextSibling();
        }
      }  
    }    
    child = child->getNextSibling();
  }

  //generate GlobalVariables file
  generateGlobalVariables(module);
  //generate main file
  generateMainFile(module, components);
  //generate pro file
  generateProFile(module, components);
}

void ComponentCodeGenerator::generateMainFile(const SFString &module, const MFString &components)
{
  //creating .h file
  SFString className(module);
  SFString hname = className + ".h";
  SFString gvClassName = module + "GlobalVariables";
    
  fstream file;
  file.open(hname);
  file.seekp(0, ios_base::end);
  
  if (file.tellp() > 0)
  {
    cerr << "error: the file " << hname << " already exists!" << endl;
    file.close();
    return;
  }  
  
  ofstream output;
  output.open(hname);      
  SFString whitespace;
  whitespace.insert(0, 48 - hname.length(), ' ');
  
  output << "////////////////////////////////////////////////////////////////////////////////" << endl;
  output << "//                            " << hname << whitespace << "//" << endl;
  output << "////////////////////////////////////////////////////////////////////////////////" << endl;
  output << endl;
  output << "#ifndef " << className.upper() << "_H" << endl;
  output << "#define " << className.upper() << "_H" << endl;
  output << endl;
  output << "#include \"" << module << "GlobalVariables.h\"" << endl;
  output << "#include \"X3DWalker.h\"" << endl;
  output << endl;
  output << "namespace X3D {" << endl;
  output << endl;
  output << "/// " << module << " module." << endl;
  output << endl;
  output << "class " << className << " : public X3DWalker" << endl;
  output << "{" << endl;
  output << "public:" << endl;
  output << "  /// Constructor." << endl;
  output << "  " << className << "();" << endl;
  output << "  /// Destructor." << endl;
  output << "  virtual ~" << className << "();" << endl;
  output << endl;
  output << "protected:" << endl;
  output << "  " << gvClassName << " *_globalVariables;" << endl;  
  output << "};" << endl;
  output << endl;
  output << "}" << endl;
  output << endl;
  output << "#endif" << endl;
  
  output.close();
  
  SFString cppname = className + ".cpp";
  output.open(cppname);
  
  output << "#include \"" << hname << "\"" << endl;  
  
  for (MFString::const_iterator it = components.begin(); it != components.end(); ++it)
    output << "#include \"" << module << (*it) << "Visitor.h\"" << endl;  
  
  output << endl;
  output << "using namespace X3D;" << endl;
  output << endl;  
  output << className << "::" << className << "()" << endl;
  output << ": " << "X3DWalker()" << endl;
  output << "{" << endl;
  output << "  // GlobalVariables assignation." << endl;
  output << "  _globalVariables = GVManager::getInstanceOf<" << gvClassName << ">();" << endl;
  output << "}" << endl;
  output << endl;
  output << className << "::~" << className << "()" << endl;
  output << "{" << endl;
  output << "}" << endl;
  output << endl;
  
  output.close();  
}

void ComponentCodeGenerator::generateProFile(const SFString &module, const MFString &components)
{
  //creating .pro file
  SFString proname = module + ".pro";
    
  fstream file;
  file.open(proname);
  file.seekp(0, ios_base::end);
  
  if (file.tellp() > 0)
  {
    cerr << "error: the file " << proname << " already exists!" << endl;
    file.close();
    return;
  }  
  
  ofstream output;
  output.open(proname);
  
  output << "TARGET = x3d" << module.lower() << endl;
  output << endl;
  output << "HEADERS = \\" << endl;
  output << "  " << module << ".h \\" << endl;
  output << "  " << module << "GlobalVariables.h \\" << endl;
  for (MFString::const_iterator it = components.begin(); it != components.end(); ++it)
    output << "  " << module << (*it) << "Visitor.h \\" << endl;

  output << endl;
  output << "SOURCES = \\" << endl;
  output << "  " << module << ".cpp \\" << endl;
  output << "  " << module << "GlobalVariables.cpp \\" << endl;  
  for (MFString::const_iterator it = components.begin(); it != components.end(); ++it)
    output << "  " << module << (*it) << "Visitor.cpp \\" << endl;
  
  output << endl;
  output << "include(../modules.pri)" << endl;
  output.close();
}

void ComponentCodeGenerator::generateGlobalVariables(const SFString &module)
{
  //creating .h file
  SFString className(module + "GlobalVariables");
  SFString hname = className + ".h";
    
  fstream file;
  file.open(hname);
  file.seekp(0, ios_base::end);
  
  if (file.tellp() > 0)
  {
    cerr << "error: the file " << hname << " already exists!" << endl;
    file.close();
    return;
  }  
  
  ofstream output;
  output.open(hname);      
  SFString whitespace;
  whitespace.insert(0, 48 - hname.length(), ' ');
  
  output << "////////////////////////////////////////////////////////////////////////////////" << endl;
  output << "//                            " << hname << whitespace << "//" << endl;
  output << "////////////////////////////////////////////////////////////////////////////////" << endl;
  output << endl;
  output << "#ifndef " << className.upper() << "_H" << endl;
  output << "#define " << className.upper() << "_H" << endl;
  output << endl;
  output << "#include \"GlobalVariables.h\"" << endl;
  output << endl;
  output << "namespace X3D {" << endl;
  output << endl;
  output << "/// GlobalVariables for the " << module << " module." << endl;
  output << endl;
  output << "class " << className << " : public GlobalVariables" << endl;
  output << "{" << endl;
  output << "public:" << endl;
  output << "  /// Constructor." << endl;
  output << "  " << className << "();" << endl;
  output << "  /// Destructor." << endl;
  output << "  virtual ~" << className << "();" << endl;
  output << "};" << endl;
  output << endl;
  output << "#endif" << endl;
  
  output.close();
  
  SFString cppname = className + ".cpp";
  output.open(cppname);
  
  output << "#include \"" << hname << "\"" << endl;  
  output << endl;
  output << "using namespace X3D;" << endl;
  output << endl;  
  output << className << "::" << className << "()" << endl;
  output << ": " << "GlobalVariables()" << endl;
  output << "{" << endl;
  output << "}" << endl;
  output << endl;
  output << className << "::~" << className << "()" << endl;
  output << "{" << endl;
  output << "}" << endl;
  output << endl;

  output.close();  
}
  
void ComponentCodeGenerator::generateFiles(const SFString &module, const SFString &component, const MFString &enterList, const MFString &walkOnList, const MFString &leaveList)
{
  //creating .h file
  SFString className(module + component + "Visitor");
  SFString hname = className + ".h";
  SFString gvClassName = module + "GlobalVariables";
    
  fstream file;
  file.open(hname);
  file.seekp(0, ios_base::end);
  
  if (file.tellp() > 0)
  {
    cerr << "error: the file " << hname << " already exists!" << endl;
    file.close();
    return;
  }  
  
  ofstream output;
  output.open(hname);      
  SFString whitespace;
  whitespace.insert(0, 48 - hname.length(), ' ');
  
  output << "////////////////////////////////////////////////////////////////////////////////" << endl;
  output << "//                            " << hname << whitespace << "//" << endl;
  output << "////////////////////////////////////////////////////////////////////////////////" << endl;
  output << endl;
  output << "#ifndef " << className.upper() << "_H" << endl;
  output << "#define " << className.upper() << "_H" << endl;
  output << endl;
  output << "#include \"" << component << "Visitor.h\"" << endl;
  output << "#include \"" << gvClassName << ".h\"" << endl;
  output << endl;
  
  //determining the list of used nodes.
  set<SFString> usedNodeList;
  for (MFString::const_iterator it = enterList.begin(); it != enterList.end(); ++it)
    usedNodeList.insert(*it);
  for (MFString::const_iterator it = walkOnList.begin(); it != walkOnList.end(); ++it)  
    usedNodeList.insert(*it);
  for (MFString::const_iterator it = leaveList.begin(); it != leaveList.end(); ++it)  
    usedNodeList.insert(*it);
    
  if (usedNodeList.size() > 0)
  {    
    for (set<SFString>::const_iterator it = usedNodeList.begin(); it != usedNodeList.end(); ++it)
      output << "class " << (*it) << ";" << endl;
    output << endl;
  }  
  output << "namespace X3D {" << endl;
  output << endl;
  output << "/// Visitor for the " << component << " component of the " << module << " module." << endl;
  output << endl;
  output << "class " << className << " : public " << component << "Visitor" << endl;
  output << "{" << endl;
  output << "public:" << endl;
  output << "  /// Constructor." << endl;
  output << "  " << className << "(NodeCreationProxy *nodeCreationProxy = 0);" << endl;
  output << "  /// Destructor." << endl;
  output << "  virtual ~" << className << "();" << endl;
  
  if (usedNodeList.size() > 0)
  {    
    output << endl;
    for (MFString::const_iterator it = enterList.begin(); it != enterList.end(); ++it)
    {
      output << "  /// Enters a " << (*it) << "." << endl;
      output << "  virtual void enter" << (*it) << "(" << (*it) << " *" << nodeFromName(*it) << ") const;" << endl;
    }
    output << endl;
    for (MFString::const_iterator it = walkOnList.begin(); it != walkOnList.end(); ++it)
    { 
      output << "  /// Walks on a " << (*it) << "." << endl;
      output << "  virtual bool walkOn" << (*it) << "(" << (*it) << " *" << nodeFromName(*it) << ", SFNode Child) const;" << endl;
    }
    output << endl;  
    for (MFString::const_iterator it = leaveList.begin(); it != leaveList.end(); ++it)  
    {
      output << "  /// Leaves a " << (*it) << "." << endl;
      output << "  virtual void leave" << (*it) << "(" << (*it) << " *" << nodeFromName(*it) << ") const;" << endl;
    }  
  }
  output << endl;
  output << "private:" << endl;
  output << "  " << gvClassName << " *_globalVariables;" << endl;  
  output << "};" << endl;
  output << endl;
  output << "#endif" << endl;
  
  output.close();
  
  SFString cppname = className + ".cpp";
  output.open(cppname);   
  output << "#include \"" << hname << "\"" << endl;
  
  if (usedNodeList.size() > 0)
  {    
    for (set<SFString>::const_iterator it = usedNodeList.begin(); it != usedNodeList.end(); ++it)
      output << "#include \"" << (*it) << ".h\"" << endl;
  }  
    
  output << endl;
  output << "using namespace X3D;" << endl;
  output << endl;  
  output << className << "::" << className << "(NodeCreationProxy *nodeCreationProxy)" << endl;
  output << ": " << component << "Visitor(nodeCreationProxy)" << endl;
  output << "{" << endl;
  
  if (usedNodeList.size() > 0)
  { 
    if (enterList.size() > 0)
    {
      output << "  // Enter functions." << endl;
      for (MFString::const_iterator it = enterList.begin(); it != enterList.end(); ++it)
        output << "  defineNewEnterFunction<" << className << ", " << (*it) << ">(&" << className << "::enter" << (*it) << ");" << endl;
      output << endl;
    }
    
    if (walkOnList.size() > 0)
    {
      output << "  // WalkOn functions." << endl;
      for (MFString::const_iterator it = walkOnList.begin(); it != walkOnList.end(); ++it)
        output << "  defineNewWalkOnFunction<" << className << ", " << (*it) << ">(&" << className << "::walkOn" << (*it) << ");" << endl;
      output << endl;
    }
    
    if (leaveList.size() > 0)
    {
      output << "  // Leave functions." << endl;
      for (MFString::const_iterator it = leaveList.begin(); it != leaveList.end(); ++it)  
        output << "  defineNewLeaveFunction<" << className << ", " << (*it) << ">(&" << className << "::leave" << (*it) << ");" << endl;
    }
  }
  output << endl;
  output << "  // GlobalVariables assignation." << endl;
  output << "  _globalVariables = GVManager::getInstanceOf<" << gvClassName << ">();" << endl;
  output << "}" << endl;
  output << endl;
  output << className << "::~" << className << "()" << endl;
  output << "{" << endl;
  output << "}" << endl;
  
  output << endl;
  for (MFString::const_iterator it = enterList.begin(); it != enterList.end(); ++it)
  {
    output << "void " << className << "::enter" << (*it) << "(" << (*it) << " *" << nodeFromName(*it) << ") const" << endl;
    output << "{" << endl;
    output << "}" << endl;
    output << endl;
  }
  
  for (MFString::const_iterator it = walkOnList.begin(); it != walkOnList.end(); ++it)
  {  
    output << "bool " << className << "::walkOn" << (*it) << "(" << (*it) << " *" << nodeFromName(*it) << ", SFNode Child) const" << endl;
    output << "{" << endl;
    output << "  return true;" << endl;
    output << "}" << endl;
    output << endl;
  }
  
  for (MFString::const_iterator it = leaveList.begin(); it != leaveList.end(); ++it)  
  {  
    output << "void " << className << "::leave" << (*it) << "(" << (*it) << " *" << nodeFromName(*it) << ") const" << endl;
    output << "{" << endl;
    output << "}" << endl;
    output << endl;
  }
  
  output.close();
}

SFString ComponentCodeGenerator::nodeFromName(const SFString &name)
{
  if (name.substr(0, 3) != "X3D")
    return name.substr(0, 1);
    
  return "N";  
}
