#include "XercesLoader.h"
#include "XercesFileElement.h"
#include "NodeCreationProxy.h"
#include "X3DAbstractNode.h"
#include "Scene.h"

#include <xercesc/util/XMLString.hpp>
#include <xercesc/sax2/DefaultHandler.hpp>
#include <xercesc/sax2/XMLReaderFactory.hpp>
#include <xercesc/util/PlatformUtils.hpp>

#include <iostream>

using namespace X3DTK;
using namespace std;
XERCES_CPP_NAMESPACE_USE

namespace X3DTK {

class SAX2X3DHandler : public DefaultHandler
{
public: 
  SAX2X3DHandler(NodeCreationProxy *nodeCreationProxy);
  ~SAX2X3DHandler();
  
  void startDocument();
  void startElement(const XMLCh* const uri, const XMLCh* const localname, const XMLCh* const qname, const XERCES_CPP_NAMESPACE_QUALIFIER Attributes &attrs);
  void endElement(const XMLCh *const uri, const XMLCh *const localname, const XMLCh *const qname);  
  void endDocument();
  void fatalError(const XERCES_CPP_NAMESPACE_QUALIFIER SAXParseException& exception);
  void error(const XERCES_CPP_NAMESPACE_QUALIFIER SAXParseException& exception);
  
  inline Scene *getScene() const {return _scene;};
  
private:
  MFNode _nodeStack;
  Scene *_scene;
  NodeCreationProxy *_ncp;
  bool _sceneDescription;
  unsigned int _depth;
};

}

X3DTK::SAX2X3DHandler::SAX2X3DHandler(NodeCreationProxy *nodeCreationProxy)
: _scene(0), _ncp(nodeCreationProxy), _sceneDescription(false), _depth(0)
{
}

X3DTK::SAX2X3DHandler::~SAX2X3DHandler()
{
}

void X3DTK::SAX2X3DHandler::startDocument()
{
  _scene = 0;
  _sceneDescription = false;
  _depth = 0;
}

void X3DTK::SAX2X3DHandler::startElement(const XMLCh* const uri, const XMLCh* const localname, const XMLCh* const qname, const Attributes &attrs)
{
  //_depth == 0, meaning that all the fathers of the node are defined nodes(have a Creation Function)
  //This variable is used to jump a sub-tree of the graph scene, when an unknown node is met.
  if (_depth == 0)
  {
    XercesFileElement e(XMLString::transcode(qname), &attrs);

    int index;  
    if (_sceneDescription)
    {
      SFNode Father = _nodeStack.back();
      SFNode N;
      index = e.getIndexAttribute("USE");
      
      if (index != -1)
      {
        N = X3DAbstractNode::DEFDict.getNodeOfName(e.getAttribute(index));
      }  
      else  
      {
        N = _ncp->createFromName(e.getName());
        if (N == 0)
        {
          _depth = 1;
          return;
        }
        N->loadAttributes(&e);
      }
      index = e.getIndexAttribute("DEF");
      if (index != -1)
        X3DAbstractNode::DEFDict.insert(e.getAttribute(index), N);

      Father->setChild(N);
      _nodeStack.push_back(N);  
    }  
  
    if (e.getName() == "Scene")
    {
      _sceneDescription = true;
      _scene = new Scene();
     _nodeStack.push_back(_scene);
    } 
  }
  else
    ++_depth;
}

void X3DTK::SAX2X3DHandler::endElement(const XMLCh *const uri, const XMLCh *const localname, const XMLCh *const qname)
{
  if (_sceneDescription)
  {
    if ((_depth == 0) && (_nodeStack.size() > 0))
      _nodeStack.pop_back();
    else
      --_depth;  
  }
}

void X3DTK::SAX2X3DHandler::endDocument()
{
  X3DAbstractNode::DEFDict.makeDEFNonAmbiguous();
  _nodeStack.clear();
}

void X3DTK::SAX2X3DHandler::error(const SAXParseException& exception)
{
  char* message = XMLString::transcode(exception.getMessage());
  cout << "Error: " << message
       << " at line: " << exception.getLineNumber()
       << endl;
}

void X3DTK::SAX2X3DHandler::fatalError(const SAXParseException& exception)
{
  char* message = XMLString::transcode(exception.getMessage());
  cout << "Fatal Error: " << message
       << " at line: " << exception.getLineNumber()
       << endl;
}

class X3DTK::XercesLoaderImplementation
{     
public:
  XercesLoaderImplementation(){};
  
  XERCES_CPP_NAMESPACE_QUALIFIER SAX2XMLReader* _parser;
  SAX2X3DHandler *_handler;
};

XercesLoader::XercesLoader(NodeCreationProxy *nodeCreationProxy)
: _impl(new XercesLoaderImplementation())
{
  ncp = nodeCreationProxy;
  
  try 
  {
    XMLPlatformUtils::Initialize();
  }
  catch (const XMLException& toCatch) 
  {
    char* message = XMLString::transcode(toCatch.getMessage());
    cout << "Error during initialization! :\n";
    cout << "Exception message is: \n" << message << "\n";
    XMLString::release(&message);
  }
  
  _impl->_parser = XMLReaderFactory::createXMLReader();
  _impl->_parser->setFeature(XMLUni::fgSAX2CoreNameSpaces, true);
  _impl->_parser->setFeature(XMLUni::fgXercesSchemaFullChecking, false);
  _impl->_parser->setFeature(XMLUni::fgSAX2CoreNameSpacePrefixes, false);
  
  _impl->_handler = new SAX2X3DHandler(nodeCreationProxy);
  _impl->_parser->setContentHandler(_impl->_handler);
  _impl->_parser->setErrorHandler(_impl->_handler);
  _impl->_parser->setDTDHandler(_impl->_handler);
}

XercesLoader::~XercesLoader()
{
  delete _impl->_parser;
  delete _impl->_handler;
}

Scene *XercesLoader::load(const char *file, bool fileValidation) const 
{
  if (fileValidation)
  {
    _impl->_parser->setFeature(XMLUni::fgXercesLoadExternalDTD, true);
    _impl->_parser->setFeature(XMLUni::fgXercesSchema, true);
    _impl->_parser->setFeature(XMLUni::fgSAX2CoreValidation, true);
  }
  else
  {
    _impl->_parser->setFeature(XMLUni::fgXercesLoadExternalDTD, false);
    _impl->_parser->setFeature(XMLUni::fgXercesSchema, false);
    _impl->_parser->setFeature(XMLUni::fgSAX2CoreValidation, false);
  }  
  
  try 
  {
    _impl->_parser->parse(file);
  }
  catch (const XMLException& toCatch) 
  {
    char* message = XMLString::transcode(toCatch.getMessage());
    cout << "Exception message is: \n"
         << message << "\n";
            XMLString::release(&message);
  }
  catch (const SAXParseException& toCatch) 
  {
    char* message = XMLString::transcode(toCatch.getMessage());
    cout << "Exception message is: \n"
         << message << "\n";
            XMLString::release(&message);
  }
  catch (...) 
  {
    cout << "Unexpected Exception \n" ;
  }
  
  Scene *scene = _impl->_handler->getScene();
  if (scene != 0)
    scene->setFileName(file);
    
  return scene;
}

